/* MODULE NAME:  k_rx_los.c
 * PURPOSE:
 *   rx los implementations for linux kernel.
 * 
 * NOTES:
 *   None.
 *
 * HISTORY
 *    8/15/2017 - Mark Hsieh, Created
 *
 * Copyright(C)      EdgeCore Corporation, 2017
 */
#include "sys_cpnt.h"
#if (TRUE == SYS_CPNT_SWDRV_SFP_RX_LOS_INTERRUPT)
#include "sys_hwcfg.h"
#include "k_rx_los.h"
#include "k_sysfun.h"

#include "drv/gpio/gpio.h"

/* linux kernel header files
 */
#include <linux/wait.h>
#include <linux/sched.h>

/* NAMING CONSTANT DECLARATIONS
 */

/* MACRO FUNCTION DECLARATIONS
 */

/* DATA TYPE DECLARATIONS
 */

/* LOCAL SUBPROGRAM DECLARATIONS
 */
static UI32_T RX_LOS_Syscall(UI32_T cmd, UI32_T arg1, UI32_T arg2, UI32_T arg3, UI32_T arg4);


static void callback0(void* param);
static void callback1(void* param);
static void callback2(void* param);
static void callback3(void* param);
static void setGpioInfo(UI32_T board_id);
static BOOL_T RX_LOS_EnableInt(void);

/* STATIC VARIABLE DECLARATIONS
 */
static wait_queue_head_t wq;
static UI32_T rx_los[4];
static UI32_T board_id = 0xff;
static BOOL_T k_rx_los_debug = FALSE;

/* TRUE for wq Ready , FALSE for wq Not Ready
 */
static BOOL_T wq_init_done = FALSE;
static BOOL_T forced_wake_up_wq = FALSE;

typedef struct GPIO_INFO_S
{
    UI32_T pin_type;
    UI32_T pin_id;
}GPIO_INFO_T;

static GPIO_INFO_T gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT];
static GPIO_INTERNAL_PIN_t pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT];
drv_gpioIsr_cb_f rx_los_callback[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT];


/* EXPORTED SUBPROGRAM BODIES
 */
/*--------------------------------------------------------------------------
 * 	FUNCTION NAME : RX_LOS_Init
 *--------------------------------------------------------------------------
 * 	PURPOSE:
 *   Initialize RX_LOS.
 *
 * 	INPUT:
 *   None.
 *
 * 	OUTPUT:
 *   None.
 *
 * 	RETURN:
 *   None.
 *
 * 	NOTES:
 *   None.
 *--------------------------------------------------------------------------
 */
void RX_LOS_Init(void)
{

}

/*--------------------------------------------------------------------------
 * FUNCTION NAME - RX_LOS_Create_InterCSC_Relation
 *--------------------------------------------------------------------------
 * PURPOSE:
 *   This function initializes all function pointer registration operations.
 *
 * INPUT:
 *   None.
 *
 * OUTPUT:
 *   None.
 *
 * RETURN
 *   None.
 *
 * NOTES:
 *   None.
 *--------------------------------------------------------------------------*/
void RX_LOS_Create_InterCSC_Relation(void)
{
    SYSFUN_RegisterCallBackFunc(SYSFUN_SYSCALL_SWDRV_SFP_RX_LOS, RX_LOS_Syscall);
}

/*--------------------------------------------------------------------------
 * 	FUNCTION NAME : RX_LOS_Syscall
 *--------------------------------------------------------------------------
 * 	PURPOSE:
 *  	This function implements the system call for RX_LOS.
 *
 * 	INPUT:
 *   cmd        --  The command to be executed
 *   arg1-arg4  --  The meaning of arg1 to arg4 depends on the cmd.
 *
 * 	OUTPUT:
 *   arg1-arg4  --  The meaning of arg1 to arg4 depends on the cmd.
 *
 * 	RETURN:
 *   The meaning of the return value depends on the cmd.
 *
 * 	NOTES:
 *   When cmd is invalid, return value is always FALSE
 *--------------------------------------------------------------------------
 */
static UI32_T RX_LOS_Syscall(UI32_T cmd, UI32_T arg1, UI32_T arg2, UI32_T arg3, UI32_T arg4)
{
#define RX_LOS_DETECTED (rx_los[0] || rx_los[1] || rx_los[2] || rx_los[3])

    BOOL_T ret = FALSE;

    switch (cmd)
    {
        case RX_LOS_TYPE_SYSCALL_CMD_ENABLE_INT:
        {
            printk("%s(%d): board_id(%lu)\n", __FUNCTION__, __LINE__, arg1);
            board_id = arg1;
            setGpioInfo(board_id);
            ret = RX_LOS_EnableInt();
            break;
        }

        case RX_LOS_TYPE_SYSCALL_CMD_WAIT_EVENT:
        {
            int rc;
            rc = wait_event_interruptible(wq, (RX_LOS_DETECTED || forced_wake_up_wq));

            if (0 != rc)
            {
                if (k_rx_los_debug)
                {
                    if (EINTR == rc)
                    {
                        printk("%s wait event error due to EINTR\n", __FUNCTION__);
                    }
                    else
                    {
                        printk("%s wait event error rc=%d\n", __FUNCTION__, rc);
                    }
                }
            }
            else
            {
                if (k_rx_los_debug)
                {
                    printk("<0>RX_LOS timestamp:%lu, %lu %lu %lu %lu forced_wake_up_wq(%d)\n",
                            SYSFUN_GetSysTick(), 
                            rx_los[0],
                            rx_los[1],
                            rx_los[2],
                            rx_los[3],
                            forced_wake_up_wq);
                }
                SYSFUN_CopyToUser((UI32_T*)arg1, &rx_los[0], sizeof(UI32_T));
                SYSFUN_CopyToUser((UI32_T*)arg2, &rx_los[1], sizeof(UI32_T));
                SYSFUN_CopyToUser((UI32_T*)arg3, &rx_los[2], sizeof(UI32_T));
                SYSFUN_CopyToUser((UI32_T*)arg4, &rx_los[3], sizeof(UI32_T));
                rx_los[0] = FALSE;
                rx_los[1] = FALSE;
                rx_los[2] = FALSE;
                rx_los[3] = FALSE;
                ret = TRUE;
            }
            forced_wake_up_wq = FALSE;
            break;
        }

        case RX_LOS_TYPE_SYSCALL_CMD_DEBUG:
        {
            k_rx_los_debug = (arg1 == 1);
            ret = TRUE;
            break;
        }

        case RX_LOS_TYPE_SYSCALL_CMD_WAKE_UP_WQ:
        {
            forced_wake_up_wq = TRUE;

            if (TRUE == wq_init_done)
            {
                wake_up(&wq);
            }
            ret = TRUE;
            break;
        }
    }
    return ret;
}

static BOOL_T RX_LOS_EnableInt(void)
{
    UI32_T id;
    int32 rc;
    gpioID gpioId;
    drv_gpio_control_t function = GPIO_CTRLFUNC_NORMAL;
    drv_gpio_direction_t direction = GPIO_DIR_IN;
    drv_gpio_interruptType_t interruptEnable = GPIO_INT_RISING_EDGE;

    rx_los_callback[0] = (drv_gpioIsr_cb_f)callback0;
    rx_los_callback[1] = (drv_gpioIsr_cb_f)callback1;
    rx_los_callback[2] = (drv_gpioIsr_cb_f)callback2;
    rx_los_callback[3] = (drv_gpioIsr_cb_f)callback3;

    for (id = 0;
         id < ARRAY_SIZE(gpio_info);
         id ++)
    {
        gpioId = GPIO_ID(gpio_info[id].pin_type, gpio_info[id].pin_id);
        rc = drv_gpio_init(gpioId, function, direction, interruptEnable);
        if (RT_ERR_OK != rc)
        {
            printk("%s(%d): rc(%d)\n", __FUNCTION__, __LINE__, rc);
            return FALSE;
        }

        rc = drv_gpio_intrHandler_register(pin[id], rx_los_callback[id]);
        if (rc != RT_ERR_OK)
        {
            printk("%s(%d): rc(%d)\n", __FUNCTION__, __LINE__, rc);
            return FALSE;
        }
    }

    init_waitqueue_head(&wq);
    wq_init_done = TRUE;
    return TRUE;
}

static void callback0(void* param)
{    
    gpioID gpioId;
    int32 rc;
    gpio_pin_data_t *gpio_data;

    gpio_data = (gpio_pin_data_t *)param;
    gpioId = GPIO_ID(gpio_info[0].pin_type, gpio_info[0].pin_id);
    rc = drv_gpio_isr_clear(gpioId);
    rx_los[0] = TRUE;
    wake_up(&wq);
    return;
}

static void callback1(void* param)
{
    gpioID gpioId;
    int32 rc;
    gpio_pin_data_t *gpio_data;

    gpio_data = (gpio_pin_data_t *)param;
    gpioId = GPIO_ID(gpio_info[1].pin_type, gpio_info[1].pin_id);
    rc = drv_gpio_isr_clear(gpioId);
    rx_los[1] = TRUE;
    wake_up(&wq);
    return;
}

static void callback2(void* param)
{
    gpioID gpioId;
    int32 rc;
    gpio_pin_data_t *gpio_data;

    gpio_data = (gpio_pin_data_t *)param;
    gpioId = GPIO_ID(gpio_info[2].pin_type, gpio_info[2].pin_id);
    rc = drv_gpio_isr_clear(gpioId);
    rx_los[2] = TRUE;
    wake_up(&wq);
    return;
}

static void callback3(void* param)
{
    gpioID gpioId;
    int32 rc;
    gpio_pin_data_t *gpio_data;

    gpio_data = (gpio_pin_data_t *)param;
    gpioId = GPIO_ID(gpio_info[3].pin_type, gpio_info[3].pin_id);
    rc = drv_gpio_isr_clear(gpioId);
    rx_los[3] = TRUE;
    wake_up(&wq);
    return;
}

static void setGpioInfo(UI32_T board_id)
{
    switch(board_id)
    {
        case 6:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_6;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_6;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }

        case 8:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_8;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_8;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }

        case 10:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_10;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_10;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }

        case 11:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_11;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_11;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }

        case 12:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_12;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_12;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }

        case 13:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_13;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_13;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }
        case 15:
        {
            GPIO_INTERNAL_PIN_t local_pin[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_INDEX_BID_15;
            GPIO_INFO_T local_gpio_info[SYS_ADPT_NBR_OF_SFP_PORT_PER_UNIT] = SYS_HWCFG_RX_LOS_GPIO_MAPPING_BID_15;
            memcpy(pin, local_pin, sizeof(pin));
            memcpy(gpio_info, local_gpio_info, sizeof(gpio_info));
            return;
        }
    }

    return;
}
#endif /* #if (TRUE == SYS_CPNT_SWDRV_SFP_RX_LOS_INTERRUPT) */

